use std::error::Error;

use component::Component;

pub mod commands;
pub mod component;
pub mod document;
pub mod errors;

type Res<T> = Result<T, Box<dyn Error>>;
type Null = Res<()>;

pub trait AsLatex {
    fn to_string(&self) -> String;
}

pub trait Populate {
    fn attach(&mut self, other: Component) -> Res<&mut Self>;
}

pub trait Opt {
    fn add_option(&mut self, opt: String);
}

#[cfg(test)]
mod tests {
    use crate::{component::*, document::*, *};
    use std::fs;
    use std::io::Write;

    #[test]
    fn primary() -> Null {
        let mut q = fs::File::create("tex/primary.tex")?;
        let mut doc = Document::new(DocumentClass::new("article"));
        doc.set_md("title", "author");
        let mut p1 = Package::new("parskip");
        p1.add_option("parfill".to_string());
        doc.new_package(p1);

        writeln!(q, "{}", doc.to_string())?;

        Ok(())
    }

    #[test]
    fn secondary() -> Null {
        // let mut q = fs::File::create("tex/secondary.tex")?;
        let mut q = fs::File::options().write(true).open("tex/secondary.tex")?;
        let mut doc = Document::new(DocumentClass::new("book"));
        doc.set_md("title", "author");
        let mut p1 = Package::new("parskip");
        p1.add_option("parfill".to_string());
        doc.new_package(p1);

        let mut part1 = Part::new("test0");
        // part1.attach(Component::Chapter(Chapter::new("testch")))?;

        let mut chap1 = Chapter::new("test1");

        let mut sec1 = Section::new("test2");

        let mut para1 = Paragraph::new();

        let mut l1 = List::new(ListType::Itemize);
        let mut l2 = List::new(ListType::Enumerate);

        let tx = TextChunk::new("This is a test sentence.", TextType::Normal);

        l1.attach(Component::Chunk(tx.clone()))?;
        l1.attach(Component::Chunk(tx.clone()))?;

        l2.attach(Component::Chunk(tx.clone()))?;
        l2.attach(Component::Chunk(tx.clone()))?;

        para1.attach(Component::List(l1.clone()))?;
        para1.attach(Component::List(l2.clone()))?;

        sec1.attach(Component::Paragraph(para1.clone()))?;
        sec1.attach(Component::Paragraph(para1.clone()))?;

        chap1.attach(Component::Section(sec1.clone()))?;
        chap1.attach(Component::Section(sec1.clone()))?;

        part1.attach(Component::Chapter(chap1.clone()))?;
        part1.attach(Component::Chapter(chap1.clone()))?;

        doc.new_component(Component::Part(part1.clone()));
        doc.new_component(Component::Part(part1.clone()));

        writeln!(q, "{}", doc.to_string())?;

        Ok(())
    }

    #[test]
    fn tertiary() -> Null {
        let mut q = fs::File::create("tex/tertiary.tex")?;
        // let mut q = fs::File::options().write(true).open("tex/tertiary.tex")?;
        let mut doc = Document::new(DocumentClass::new("book"));
        doc.set_md("title", "author");
        let mut p1 = Package::new("parskip");
        p1.add_option("parfill".to_string());
        doc.new_package(p1);

        let mut p1 = Part::new("one");
        p1.attach(Component::Chapter(Chapter::new("c1")))?.attach(Component::Chapter(Chapter::new("c2")))?;

        doc.new_component(Component::Part(p1));

        writeln!(q, "{}", doc.to_string())?;

        Ok(())
    }
}
